//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// Nes Base Mapper - Implementation of the NesBaseMapper class (mapper #0)
//////////////////////////////////////////////////////////////////////

#include "NesBaseMapper.h"

//////////////////////////////////////////////////////////////////////
// Procedure: NesBaseMapper constructor (special)
//////////////////////////////////////////////////////////////////////
// Description: This procedure links the various NES components
// to itsself.
//////////////////////////////////////////////////////////////////////
NesBaseMapper::NesBaseMapper(NesCartridge *linkCartridge, NesCPU *linkCPU, NesPPU *linkPPU)
{
	mCartridge = linkCartridge;
	mCpu = linkCPU;
	mPpu = linkPPU;
	
	bytPRGCount = mCartridge->header.PRGROM_count;
	bytCHRCount = mCartridge->header.CHRROM_count;
	
	initMaps();
}




//////////////////////////////////////////////////////////////////////
// Procedure: initMemory
//////////////////////////////////////////////////////////////////////
// Description: This procedure loads the proper banks of PRGROM
// and CHRROM from a cartridge into memory on a POWER ON or RESET.
// For mapper #0, the PRG-ROM pages should all fit into CPU memory and
// the 8k CHRROM goes into the PPU memory
//////////////////////////////////////////////////////////////////////
void NesBaseMapper::initMemory() {			
	if (bytPRGCount > 0) {				
		if (bytPRGCount == 1) {
			swapPRGBank16(0x8000, 0);
			swapPRGBank16(0xC000, 0);				
		} else {
			swapPRGBank32(0);	
		}		
	}
		
	// Copy cartridge CHRRAM contents into NES PPU memory
	if (bytCHRCount > 0) {			
		swapCHRBank8(0);
	}	
}


// Initializes the bank maps which are used for optimizing/only swapping when needed
void NesBaseMapper::initMaps() {
	PrgMap32k[0] = -1;
	PrgMap16k[0] = PrgMap16k[1] = -1;
	PrgMap8k[0] = PrgMap8k[1] = PrgMap8k[2] = PrgMap8k[3] = -1;

	ChrMap8k[0] = -1;
	ChrMap4k[0] = ChrMap4k[1] = -1;
	
	for (int i = 0; i < 8; i++) {
		if (i < 4) ChrMap2k[i] = -1;
		ChrMap1k[i] = -1;
	}
}

//////////////////////////////////////////////////////////////////////
// Procedure: write
//////////////////////////////////////////////////////////////////////
// Description: This procedure accepts a command from the CPU.  Commands
// are issued by trying to write to PRGROM (i.e. sta $8000)  Mapper #0
// is simple and does not accept any commands, but more advanced mappers 
// do and this is a crucial procedure to implement for those cases.
//////////////////////////////////////////////////////////////////////
void NesBaseMapper::write(Uint16 addr, Uint8 data)
{
	// Base Mapper/Mapper #0: Do nothing
}


////////////////////////////////////////////////////////////////////
// Provide some commonly-used swapping routines for all mappers
////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////
// Swap a 8K bank of PRGROM into CPU memory
// Slightly confusing because the PRGROM is stored in 16K increments
////////////////////////////////////////////////////////////////////
void NesBaseMapper::swapPRGBank8(Uint16 intOffset, Uint8 bytBank)
{
	int bankSlot = (intOffset & 0x6000) >> 13;
	
	if (PrgMap8k[bankSlot] != bytBank) {
		int intBankOffset = 0x0000, intPRGBank16 = bytBank >> 1;
		
		// If the bank is the upper portion of a 16K page then split it
		if (bytBank & 0x01) intBankOffset = 0x2000;
		
		memcpy( (void *)&mCpu->memory[intOffset], (void *)&mCartridge->PRGROM_Pages[intPRGBank16][intBankOffset], 0x2000);
		/*int intByteCtr = 0;
		while (intByteCtr < 0x2000)
		{
			mCpu->memory[intByteCtr + intOffset] = mCartridge->PRGROM_Pages[intPRGBank16][intByteCtr + intBankOffset];
			intByteCtr++;
		}
		*/
		
		PrgMap8k[bankSlot] = bytBank;
	}	
}


////////////////////////////////////////////////////////////////////
// Swap a 16k bank of PRGROM into CPU memory
// For example:  to load the 1st 16k of PRGROM into 8000 : swapPRGBank16(0x8000,0);
////////////////////////////////////////////////////////////////////
void NesBaseMapper::swapPRGBank16(Uint16 intOffset, Uint8 bytBank)
{
	int bankSlot = (intOffset & 0x4000) >> 14;
	//int bankSlot = 0x00;
	//if (intOffset == 0xC000) bankSlot = 0x01;
	
	if (PrgMap16k[bankSlot] != bytBank) {
		memcpy( (void *)&mCpu->memory[intOffset], (void *)&mCartridge->PRGROM_Pages[bytBank][0], 0x4000);
		
		/*
		int intByteCtr = 0;
		while (intByteCtr < 0x4000)
		{
			mCpu->memory[intByteCtr + intOffset] = mCartridge->PRGROM_Pages[bytBank][intByteCtr];
			intByteCtr++;
		}
		*/
		PrgMap16k[bankSlot] = bytBank;
	}
	
	int eb = (intOffset == 0x8000) ? 0 : 2;
	PrgMap8k[ eb     ] = (bytBank << 1);
	PrgMap8k[ eb + 1 ] = (bytBank << 1) + 1;
}


////////////////////////////////////////////////////////////////////
// Swap a 32k bank of PRGROM into CPU memory.  Since CPU memory is 32k, no need to pass offset
// parameter
////////////////////////////////////////////////////////////////////
void NesBaseMapper::swapPRGBank32(Uint8 bytBank)
{
	if (PrgMap32k[0] != bytBank) {
		PrgMap32k[0] = bytBank;

		// Interpret bytBank paramter as the bytBankth 32k page to swap in
		// Need to multiply by 2 to get the 16k page
		bytBank <<= 1;
		memcpy( (void *)&mCpu->memory[0x8000], (void *)&mCartridge->PRGROM_Pages[bytBank][0], 0x4000);
		memcpy( (void *)&mCpu->memory[0xC000], (void *)&mCartridge->PRGROM_Pages[bytBank+1][0], 0x4000);
		/*
		int intByteCtr = 0;
		while (intByteCtr < 0x4000)
		{
			mCpu->memory[intByteCtr + 0x8000] = mCartridge->PRGROM_Pages[bytBank][intByteCtr];
			mCpu->memory[intByteCtr + 0xC000] = mCartridge->PRGROM_Pages[bytBank + 1][intByteCtr];
			intByteCtr++;
		}
		*/		
	}
	PrgMap8k[ 0 ] = (bytBank << 2);
	PrgMap8k[ 1 ] = (bytBank << 2) + 1;
	PrgMap8k[ 2 ] = (bytBank << 2) + 2;
	PrgMap8k[ 3 ] = (bytBank << 2) + 3;
}



////////////////////////////////////////////////////////////////////
// Swap a 1k bank of CHRROM into PPU memory
////////////////////////////////////////////////////////////////////
void NesBaseMapper::swapCHRBank1(Uint16 intOffset, Uint8 bytBank) {
	if (intOffset < 0x2000) {
		int bankSlot = (intOffset & 0x1C00) >> 10;
		if (ChrMap1k[bankSlot] != bytBank) {
			// Divide bank by 8 to get the actual page of data to look into
			int intBank = bytBank >> 3;

			// Mod by 8 to get the page offset to look into
			// TODO: remove this multiplication with a shift
			int intBankOffset = (bytBank & 0x07) * 0x400;

			memcpy((void *)&mPpu->VRAM->memory[intOffset], (void *)&mCartridge->CHRROM_Pages[intBank][intBankOffset], 0x400);

			ChrMap1k[bankSlot] = bytBank;
		}
	} else {
		// Some mappers (#19) swap in the name table data, too...
		int intBank = bytBank >> 3;
		int intBankOffset = (bytBank & 0x07) * 0x400;

		int intNT = (intOffset & 0xC00) >> 10;
		memcpy(&mPpu->VRAM->bytNameTable[intNT][0], &mCartridge->CHRROM_Pages[intBank][intBankOffset], 0x400);
	}
}


////////////////////////////////////////////////////////////////////
// Swap a 2k bank of CHRROM into PPU memory
////////////////////////////////////////////////////////////////////
void NesBaseMapper::swapCHRBank2(Uint16 intOffset, Uint8 bytBank) {
	// If this bank is not already mapped here, then swap it in	
	if (ChrMap2k[ (intOffset & 0x1800) >> 11 ] != bytBank ) {
		ChrMap2k[ (intOffset & 0x1800) >> 11 ] = bytBank;

		// Bank parameter is interpreted as "the bytBank-th 2k data bank", so we might have 
		// to jump into the middle of a 2k page (CHRROM is stored in cartridge in 8K increments)
		int intBankOffset = (bytBank & 0x03) << 11;
		//if (bytBank & 0x01) intBankOffset = 0x1000;
		
		// CHRROM stored in 8K increments, so we divide our 2k bank by 4 to get the 8k offset
		bytBank >>= 2;
			
		memcpy((void *)&mPpu->VRAM->memory[intOffset], (void *)&mCartridge->CHRROM_Pages[bytBank][intBankOffset], 0x800);		
	}
}


////////////////////////////////////////////////////////////////////
// Swap a 4k bank of CHRROM into PPU memory
////////////////////////////////////////////////////////////////////
void NesBaseMapper::swapCHRBank4(Uint16 intOffset, Uint8 bytBank)
{
	// For safety, make sure bank is in valid range
	//bytBank = bytBank % (mCartridge->header.CHRROM_count << 1);

	// If this bank is not already mapped here, then swap it in	
	if (ChrMap4k[ (intOffset & 0x1000) >> 12 ] != bytBank ) {
		ChrMap4k[ (intOffset & 0x1000) >> 12 ] = bytBank;

		// Bank parameter is interpreted as "the bytBank-th 4k data bank", so we might have 
		// to jump into the middle of a 8k page (CHRROM is stored in cartridge in 8K increments)
		int intBankOffset = 0;
		if (bytBank & 0x01) intBankOffset = 0x1000;
		
		// CHRROM stored in 8K increments, so we divide by 2 to get 4k offset		
		bytBank >>= 1;
			
		memcpy((void *)&mPpu->VRAM->memory[intOffset], (void *)&mCartridge->CHRROM_Pages[bytBank][intBankOffset], 0x1000);		
	}
}

////////////////////////////////////////////////////////////////////
// Swap a 8k bank of CHRROM into PPU memory.  Since the PPU CHRROM area is only 8k
// there is no need for an offset paramter
////////////////////////////////////////////////////////////////////
void NesBaseMapper::swapCHRBank8(Uint8 bytBank)
{	
	if (ChrMap8k[0] != bytBank) {
		memcpy((void *)&mPpu->VRAM->memory[0], (void *)&mCartridge->CHRROM_Pages[bytBank][0], 0x2000);		
		/*int intByteCtr = 0;
		while (intByteCtr < 0x2000)
		{
			mPpu->VRAM->memory[intByteCtr] = mCartridge->CHRROM_Pages[bytBank][intByteCtr];				
			intByteCtr++;
		}*/
		
		ChrMap8k[0] = bytBank;
	}
}



//-------------------------------------------------------------------
// Procedure: saveState
//-------------------------------------------------------------------
// Description: Saves the state of the cartridge mapper to the file stream
//-------------------------------------------------------------------
void NesBaseMapper::saveState(FILE *stream) {
	fwrite(&PrgMap8k[0], sizeof(int)*4, 1, stream);	// PRGROM map
	fwrite(&PrgMap16k[0], sizeof(int)*2, 1, stream);// PRGROM map
	fwrite(&PrgMap32k[0], sizeof(int)*1, 1, stream);// PRGROM map
	
	fwrite(&ChrMap1k[0], sizeof(int)*8, 1, stream);// CHRRAM map
	fwrite(&ChrMap4k[0], sizeof(int)*2, 1, stream);// CHRRAM map
	fwrite(&ChrMap8k[0], sizeof(int)*1, 1, stream);// CHRRAM map		
}




//-------------------------------------------------------------------
// Procedure: loadState
//-------------------------------------------------------------------
// Description: Loads the state of the cartridge mapper from the file stream
//-------------------------------------------------------------------
void NesBaseMapper::loadState(FILE *stream) {	
	fread(&PrgMap8k[0], sizeof(int)*4, 1, stream);  // PRGROM map
	fread(&PrgMap16k[0], sizeof(int)*2, 1, stream); // PRGROM map
	fread(&PrgMap32k[0], sizeof(int)*1, 1, stream); // PRGROM map
	
	fread(&ChrMap1k[0], sizeof(int)*8, 1, stream);// CHRRAM map
	fread(&ChrMap4k[0], sizeof(int)*2, 1, stream);// CHRRAM map
	fread(&ChrMap8k[0], sizeof(int)*1, 1, stream);// CHRRAM map		

	// Swap in the PRGROM 8k, since that is guaranteed to be valid
	for (int i = 0; i < 4; i++) {
		// Forcibly swap in the PRGROM; we can't use the swap method because we
		// loaded the 8k map and it will think the PRGROM is already swapped in (was a big bug!)
		int intBankOffset = 0x0000, intPRGBank16 = PrgMap8k[i] >> 1;
		if (PrgMap8k[i] & 0x01) intBankOffset = 0x2000;		
		memcpy( (void *)&mCpu->memory[0x8000 + i * 0x2000], (void *)&mCartridge->PRGROM_Pages[intPRGBank16][intBankOffset], 0x2000);		
	}
}